#include "psublast/libc.h"
#include "psublast/types.h"
#include "psublast/edit.h"
#include "psublast/args.h"
#include "psublast/seq.h"
#include "psublast/dna.h"
#include "psublast/misc.h"
#include "psublast/prnt.h"

#ifndef __lint
static const char rcsid[] =
"$Id: prnt.c,v 1.23 1999/09/23 21:32:08 schwartz Exp schwartz $";
#endif

/* XXX */
static int offset1;
static int offset2;

enum { BUFSIZE=128 };
#define ckprintf (void)printf /* XXX */

/* print_align_header  -------------  print the top part of an alignment file */
void print_align_header(SEQ *seq1, SEQ *seq2, argv_scores_t *ds)
{
	print_align_header_n(seq1, seq2, ds, 0);
}

static char *subseq_label(char *buf, unsigned int size, int n)
{
	assert(size > 0);
	buf[0] = 0;
	if (n > 0) snprintf(buf, size, " (subsequence #%d)", n);
	return buf;
}

static const char* revflag(SEQ *s)
{
	return (s->flags & SEQ_IS_REVCOMP) ? "-" : "";
}

static const char* revlabel(SEQ *s)
{
	return (s->flags & SEQ_IS_REVCOMP) ? " (reverse complement)" : "";
}

void print_align_header_n(SEQ *seq1, SEQ *seq2, argv_scores_t *ds, int n)
{
	int f, t, F, T;
	char buf[BUFSIZE];

	ckprintf("#:lav\n\nd {\n  \"");
	ck_argc("print_align_header");
	fprintf_argv(stdout);
	if (ds->M) {
		ckprintf("\n   M = %d, I = %d, V = %d", ds->M, ds->I, ds->V);
		if (ds->O) ckprintf(", O = %d, E = %g", ds->O, ds->E);
	} else {
		ckprintf("\n   Blosum62 scores");
	}
	ckprintf("\"\n}\n");

	if (get_argval('f', &f)) {
		if (!get_argval('t',&t) || !get_argval('F',&F) ||
		    !get_argval('T',&T))
			fatal("Inconsistent use of `f`, `t`, `F', `T' args.");
		offset1 = SEQ_FROM(seq1) - f;
		offset2 = SEQ_FROM(seq2) - F;
	} else {
		f = SEQ_FROM(seq1); t = SEQ_TO(seq1);
		F = SEQ_FROM(seq2); T = SEQ_TO(seq2);
		offset1 = offset2 = 0;
	}
	ckprintf("s {\n  \"%s%s\" %d %d\n  \"%s%s\" %d %d\n}\n",
			SEQ_NAME(seq1), revflag(seq1), f, t,
			SEQ_NAME(seq2), revflag(seq2), F, T);
	ckprintf("h {\n   \"%s%s\"\n   \"%s%s%s\"\n}\n", 
		SEQ_HEAD(seq1), 
		revlabel(seq1),
		SEQ_HEAD(seq2),
		revlabel(seq2),
		subseq_label(buf, sizeof buf, n));
}

/* print_block  ---------------------------------- print a gap-free alignment */
void print_block(int score, int beg1, int end1, int beg2, int end2, int f, int pm)
{
	beg1 += offset1;
	end1 += offset1;
	beg2 += offset2;
	end2 += offset2;

	ckprintf("a {\n  s %d\n  b %d %d\n  e %d %d\n  f %d\n",
		score, beg1, beg2, end1, end2, f);
	ckprintf("  l %d %d %d %d %d\n}\n", beg1, beg2, end1, end2, pm);
}

/* print_align  ----------------------------------- print a general alignment */
/* XXX - is this still used anywhere?? */
void print_align(int score, uchar *seq1, uchar *seq2, int beg1, int end1, int beg2, int end2,int *S)
{
	int M, N, i, j, op, start_i, start_j, match, run, pct;
	uchar *P, *p, *q;
	
	beg1 += offset1;
  	end1 += offset1;
	beg2 += offset2;
	end2 += offset2;

	M = end1 - beg1 + 1;
	N = end2 - beg2 + 1;
	ckprintf("a {\n  s %d\n  b %d %d\n  e %d %d\n",
		score, beg1, beg2, end1, end2);
	for (i = j = 0; i < M || j < N; ) {
		start_i = i;
		start_j = j;
		match = 0;
		P= p = seq1 + beg1 + i - 1;
		q = seq2 + beg2 + j - 1;
		while (i < M && j < N && *S == 0) {
			if (*p++ == *q++)
				++match;
			++i;
			++j;
			++S;
		}
		run = p - P;
		pct = (run > 0) ? ((100*match + run/2)/run) : 0; /* round */
		ckprintf("  l %d %d %d %d %d\n",
			beg1+start_i, beg2+start_j, beg1+i-1, beg2+j-1, pct);
		if (i < M || j < N) {
			if ((op = *S++) > 0)
				j += op;
			else
				i -= op;
		}
    	}
	ckprintf("}\n");
}

static void print_align_lav_fn(
    const uchar *seq1, const uchar *seq2,
    int beg1, int end1,
    int beg2, int end2,
    edit_script_t *S,
    int score,
    void (*prefn)(int b1, int b2, int e1, int e2, int s),
    void (*perfn)(int b1, int b2, int e1, int e2, int p),
    void (*postfn)(int t, int m, int n, int l)
)
{
        int M, N, i, j, o, nG = 0, lG = 0;
	int total_run = 0, total_match = 0;
 
        beg1 += offset1;
        end1 += offset1;
        beg2 += offset2;
        end2 += offset2;
 
        M = end1 - beg1 + 1;
        N = end2 - beg2 + 1;

	prefn(beg1, beg2, end1, end2, score);
        for (o = i = j = 0; i < M || j < N; ) {
                int start_i = i;
                int start_j = j; 
                int match = 0;
		int run;

	        run = es_rep_len(S, &o, seq1+beg1+i-1, seq2+beg2+j-1, &match);
		i += run; j += run;
		total_match += match;
		total_run += run;

		perfn(beg1+start_i, beg2+start_j, beg1+i-1, beg2+j-1, 
		      align_match_percent(run, match));

                if (i < M || j < N) {
		  nG++;
		  lG += es_indel_len(S, &o, &i, &j);
		}
		 
        }
	postfn(total_run, total_match, nG, lG);
}

static void prefn_lav(int b1, int b2, int e1, int e2, int s)
{
    ckprintf("a {\n  s %d\n  b %d %d\n  e %d %d\n", s, b1, b2, e1, e2);
}
static void perfn_lav(int b1, int b2, int e1, int e2, int p)
{
    ckprintf("  l %d %d %d %d %d\n", b1, b2, e1, e2, p);
}
/*ARGSUSED*/
static void postfn_lav(int t, int m, int n, int l)
{
    ckprintf("}\n");
}

/*ARGSUSED*/
static void prefn_sum(int b1, int b2, int e1, int e2, int s) { /* */ }
/*ARGSUSED*/
static void perfn_sum(int b1, int b2, int e1, int e2, int p) { /* */ }
static void postfn_sum(int t, int m, int n, int l)
{
    /* match, mismatch, #Gaps, total gap length */
    ckprintf("%d %d %d %d\n", m, t-m, n, l);
}

void print_align_lav(int score, const uchar *seq1, const uchar *seq2,
 int beg1, int end1, int beg2, int end2, edit_script_t *S)
{

    print_align_lav_fn(seq1, seq2, beg1, end1, beg2, end2, S, score,
        prefn_lav, perfn_lav, postfn_lav);
}

void print_align_summary(int score, const uchar *seq1, const uchar *seq2,
 int beg1, int end1, int beg2, int end2, edit_script_t *S)
{
 
    print_align_lav_fn(seq1, seq2, beg1, end1, beg2, end2, S, score,
        prefn_sum, perfn_sum, postfn_sum);
}

int align_match_percent(int run, int match)
{
	return (run > 0) ? ((100*match + run/2)/run) : 0; /* round */
}

void print_align_list(align_t *a) 
{
	while (a) {
		print_align_lav(a->score, a->seq1, a->seq2, a->beg1, a->end1,
			a->beg2, a->end2, a->script);
		a = a->next_align;
	}
}

void free_align_list(align_t *a)
{
	align_t *b;

	while (a) {
		edit_script_free(a->script);
		b = a->next_align;
		free(a);
		a = b;
	}
}
